# Copyright 2020 LINE Corp.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import select
import signal
import socket
import sys
import threading
from wsgiref.simple_server import make_server

from oslo_config import cfg
from oslo_log import log as logging
from prometheus_client import make_wsgi_app

from oslo_metrics import message_router


oslo_metrics_configs = [
    cfg.StrOpt('metrics_socket_file',
               default='/var/tmp/metrics_collector.sock',  # nosec
               help='Unix domain socket file to be used'
                    ' to send rpc related metrics'),
    cfg.PortOpt('prometheus_port', default=3000,
                help='Port number to expose metrics in prometheus format.'),
]
cfg.CONF.register_opts(oslo_metrics_configs, group='oslo_metrics')


LOG = logging.getLogger(__name__)
CONF = cfg.CONF
logging.register_options(CONF)
logging.setup(CONF, 'oslo-metrics')


class MetricsListener():

    def __init__(self, socket_path):
        self.socket_path = socket_path
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        self.unlink(socket_path)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.socket_path)
        self.start = True
        self.router = message_router.MessageRouter()

    def unlink(self, socket_path):
        try:
            os.unlink(socket_path)
        except OSError:
            if os.path.exists(socket_path):
                raise

    def serve(self):
        while self.start:
            readable, writable, exceptional = select.select(
                [self.socket], [], [], 1)
            if len(readable) == 0:
                continue
            try:
                LOG.debug("wait for socket.recv")
                # 1 message size should be smaller than 65565
                msg = self.socket.recv(65565)
                LOG.debug("got message")
                self.router.process(msg)
            except socket.timeout:
                pass

    def stop(self):
        self.socket.close()
        self.start = False


httpd = None


def handle_sigterm(_signum, _frame):
    LOG.debug("Caught sigterm")
    shutdown_thread = threading.Thread(target=httpd.shutdown)
    shutdown_thread.start()


def main():
    cfg.CONF(sys.argv[1:])
    socket_path = cfg.CONF.oslo_metrics.metrics_socket_file
    m = MetricsListener(socket_path)
    try:
        os.chmod(socket_path, 0o660)  # nosec
    except OSError:
        LOG.error("Changing the mode of the file failed.... continuing")
    mt = threading.Thread(target=m.serve)
    LOG.info("Start oslo.metrics")
    mt.start()

    app = make_wsgi_app()
    try:
        global httpd
        httpd = make_server('', CONF.oslo_metrics.prometheus_port, app)
        signal.signal(signal.SIGTERM, handle_sigterm)
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    finally:
        LOG.info("Try to stop...")
        os.remove(cfg.CONF.oslo_metrics.metrics_socket_file)
        m.stop()
        httpd.server_close()


if __name__ == "__main__":
    main()



# Copyright 2020 LINE Corp.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log as logging
from oslo_utils import importutils

from oslo_metrics import message_type

LOG = logging.getLogger(__name__)


MODULE_LISTS = [
    "oslo_metrics.metrics.oslo_messaging",
]


class MessageRouter():

    def __init__(self):
        self.modules = {}
        for m_str in MODULE_LISTS:
            mod = importutils.try_import(m_str, False)
            if not mod:
                LOG.error("Failed to load module %s" % m_str)
            self.modules[m_str.split('.')[-1]] = mod

    def process(self, raw_string):
        try:
            metric = message_type.Metric.from_json(raw_string.decode())
            self.dispatch(metric)
        except Exception as e:
            LOG.error("Failed to parse: %s", e)

    def dispatch(self, metric):
        if metric.module not in self.modules:
            LOG.error("Failed to lookup modules by %s" % metric.module)
            return
        mod = self.modules.get(metric.module)

        # Get metric
        try:
            metric_definition = getattr(mod, metric.name)
        except AttributeError as e:
            LOG.error("Failed to load metrics {}: {}".format(metric.name, e))
            return

        # Get labels
        try:
            metric_with_label = getattr(metric_definition, "labels")
            metric_with_label = metric_with_label(**metric.labels)
        except AttributeError as e:
            LOG.error("Failed to load labels func from metrics %s: %s" %
                      (metric.name, e))
            return
        LOG.debug("Get labels with {}: {}".format(metric.name, metric.labels))

        # perform action
        try:
            embed_action = getattr(metric_with_label, metric.action.action)
            if metric.action.value is not None:
                embed_action(metric.action.value)
            else:
                embed_action()
        except AttributeError as e:
            LOG.error("Failed to perform metric actionv %s, %s: %s" %
                      (metric.action.action, metric.action.value, e))
            return
        LOG.debug("Perform action %s for %s metrics" %
                  (metric.action.action, metric.name))



# Copyright 2020 LINE Corp.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import prometheus_client

rpc_server_common_labels = [
    'exchange', 'topic', 'server', 'endpoint', 'namespace',
    'version', 'method', 'process'
]
rpc_client_common_labels = [
    'call_type', 'exchange', 'topic', 'namespace', 'version',
    'server', 'fanout', 'process', 'method', 'timeout'
]

rpc_processing_seconds_buckets = [
    0.01, 0.025, 0.05, 0.075, 0.1, 0.25, 0.5, 0.75, 1.0,
    2.5, 5.0, 7.5, 10.0, 25.0, 50.0, 75.0, 100
]

# RPC Server Metrics
rpc_server_invocation_start_total = prometheus_client.Counter(
    'oslo_messaging_rpc_server_invocation_start_total',
    'Total number of RPC invocation start. This doesn\'t count'
    'if rpc server failed to find method from endpoints.',
    rpc_server_common_labels)

rpc_server_invocation_end_total = prometheus_client.Counter(
    'oslo_messaging_rpc_server_invocation_end_total',
    'Total number of RPC invocation end.',
    rpc_server_common_labels)

rpc_server_processing_seconds = prometheus_client.Histogram(
    'oslo_messaging_rpc_server_processing_seconds',
    'Duration of RPC processing.',
    rpc_server_common_labels,
    buckets=rpc_processing_seconds_buckets)

rpc_server_exception_total = prometheus_client.Counter(
    'oslo_messaging_rpc_server_exception_total',
    'Total number of exception while RPC processing.',
    rpc_server_common_labels + ['exception'])

# RPC Client Metrics
rpc_client_invocation_start_total = prometheus_client.Counter(
    'oslo_messaging_rpc_client_invocation_start_total',
    'Total number of RPC invocation start.',
    rpc_client_common_labels)

rpc_client_invocation_end_total = prometheus_client.Counter(
    'oslo_messaging_rpc_client_invocation_end_total',
    'Total number of RPC invocation end.',
    rpc_client_common_labels)

rpc_client_processing_seconds = prometheus_client.Histogram(
    'oslo_messaging_rpc_client_processing_seconds',
    'Duration of RPC processing.',
    rpc_client_common_labels,
    buckets=rpc_processing_seconds_buckets)

rpc_client_exception_total = prometheus_client.Counter(
    'oslo_messaging_rpc_client_exception_total',
    'Total number of exception while RPC processing.',
    rpc_client_common_labels + ['exception', ])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_message_process
--------------------
Check that messages are processed correctly
"""

from unittest import mock

from oslo_metrics import message_router
from oslotest import base
import prometheus_client


class TestProcessMessage(base.BaseTestCase):

    def setUp(self):
        super().setUp()

    def test_process_counter(self):
        received_json = b"""{
  "module": "oslo_messaging",
  "name": "rpc_server_invocation_start_total",
  "action": {
    "action": "inc",
    "value": null
  },
  "labels": {
    "exchange": "foo",
    "topic": "bar",
    "server": "foobar",
    "endpoint": "endpoint",
    "namespace": "ns",
    "version": "v2",
    "method": "get",
    "process": "done"
  }
}"""

        with mock.patch.object(
            prometheus_client.Counter, 'inc',
        ) as mock_inc:
            router = message_router.MessageRouter()
            router.process(received_json)
            mock_inc.assert_called_once_with()

    def test_process_histogram(self):
        received_json = b"""{
  "module": "oslo_messaging",
  "name": "rpc_client_processing_seconds",
  "action": {
    "action": "observe",
    "value": 1.26
  },
  "labels": {
    "call_type": "call",
    "exchange": "foo",
    "topic": "bar",
    "method": "get",
    "server": "foobar",
    "namespace": "ns",
    "version": "v2",
    "process": "done",
    "fanout": "foo",
    "timeout": 10
  }
}"""

        with mock.patch.object(
            prometheus_client.Histogram, 'observe',
        ) as mock_inc:
            router = message_router.MessageRouter()
            router.process(received_json)
            mock_inc.assert_called_once_with(1.26)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_message_validation
--------------------
Check that messages validation is working properly
"""

import json
from oslo_metrics import message_type
from oslotest import base


class TestMetricValidation(base.BaseTestCase):
    def setUp(self):
        super().setUp()

    def assertRaisesWithMessage(self, message, func, *args, **kwargs):
        try:
            func(*args, **kwargs)
            self.assertFail()
        except Exception as e:
            self.assertEqual(message, e.message)

    def test_message_validation(self):
        metric = dict()
        message = "module should be specified"
        self.assertRaisesWithMessage(
            message, message_type.Metric.from_json, json.dumps(metric))

        metric['module'] = "test"
        message = "name should be specified"
        self.assertRaisesWithMessage(
            message, message_type.Metric.from_json, json.dumps(metric))

        metric['name'] = "test"
        message = "action should be specified"
        self.assertRaisesWithMessage(
            message, message_type.Metric.from_json, json.dumps(metric))

        metric['action'] = "test"
        message = "labels should be specified"
        self.assertRaisesWithMessage(
            message, message_type.Metric.from_json, json.dumps(metric))

        metric['labels'] = "test_label"
        message = "action need 'value' field"
        self.assertRaisesWithMessage(
            message, message_type.Metric.from_json, json.dumps(metric))

        metric['action'] = {"value": "1"}
        message = "action need 'action' field"
        self.assertRaisesWithMessage(
            message, message_type.Metric.from_json, json.dumps(metric))

        metric['action']['action'] = "test"
        message = "action should be choosen from ['inc', 'observe']"
        self.assertRaisesWithMessage(
            message, message_type.Metric.from_json, json.dumps(metric))
